from frasco.ext import *
from redis import StrictRedis
from werkzeug.local import LocalProxy
from .templating import CacheFragmentExtension


class FrascoRedis(Extension):
    name = "frasco_redis"
    defaults = {"url": "redis://localhost:6379/0",
                "fragment_cache_timeout": 3600}

    def _init_app(self, app, state):
        state.connection = StrictRedis.from_url(state.options["url"])
        app.jinja_env.add_extension(CacheFragmentExtension)


def get_current_redis():
    return get_extension_state('frasco_redis').connection

redis = LocalProxy(get_current_redis)



from flask import json, make_response, Blueprint, jsonify, g
from frasco.helpers import url_for
from frasco.utils import cached_property, join_url_rule
from frasco.request_params import disable_request_params
from frasco.marshaller import disable_marshaller
from werkzeug.wrappers import Response
import functools
from .swagger import build_swagger_spec, build_swagger_client
from .errors import *


def wrap_service_func_for_endpoint(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            rv = func(*args, **kwargs)
        except ApiError as e:
            return make_response(json.dumps({"error": unicode(e)}), e.http_code)
        if isinstance(rv, Response):
            return rv
        return json.dumps(rv), {'Content-Type': 'application/json;charset=UTF-8'}
    return wrapper


def wrap_service_func_for_internal_call(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        with disable_marshaller(), disable_request_params():
            return func(*args, **kwargs)
    return wrapper


class ApiService(object):
    def __init__(self, name=None, url_prefix=None):
        self.name = name
        self.url_prefix = url_prefix
        self.rules = []
        self.endpoint_funcs = {}
        self.decorators = []

    def add_endpoint(self, rule, func, **options):
        endpoint = options.pop("endpoint", func.__name__)
        self.rules.append((rule, endpoint, options))
        self.endpoint_funcs[endpoint] = self._apply_decorators(func)

    def iter_endpoints(self):
        for rule, endpoint, options in self.rules:
            yield rule, endpoint, self.endpoint_funcs[endpoint], options

    def route(self, rule, **options):
        def decorator(func):
            self.add_endpoint(rule, func, **options)
            return func
        return decorator

    def GET(self, *args, **kwargs):
        kwargs['methods'] = ['GET']
        return self.route(*args, **kwargs)

    def POST(self, *args, **kwargs):
        kwargs['methods'] = ['POST']
        return self.route(*args, **kwargs)

    def PUT(self, *args, **kwargs):
        kwargs['methods'] = ['PUT']
        return self.route(*args, **kwargs)

    def DELETE(self, *args, **kwargs):
        kwargs['methods'] = ['DELETE']
        return self.route(*args, **kwargs)

    def __getattr__(self, name):
        if name not in self.endpoint_funcs:
            raise ValueError("Endpoint '%s' does not exist in service '%s'" % (name, self.name))
        return wrap_service_func_for_internal_call(self.endpoint_funcs[name])

    def mixin(self, service, decorator=None, url_prefix=None):
        for rule, endpoint, func, options in service.iter_endpoints():
            rule = join_url_rule(url_prefix, rule)
            if decorator:
                func = decorator(func)
            self.add_endpoint(rule, func, endpoint=endpoint, **options)

    def decorator(self, func):
        self.decorators.append(func)
        return func

    def before_request(self, before_request_func):
        def decorator(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                before_request_func()
                return func(*args, **kwargs)
            return wrapper
        self.decorators.append(decorator)
        return before_request_func

    def _apply_decorators(self, func):
        for decorator in reversed(self.decorators):
            func = decorator(func)
        return func


class ApiVersion(ApiService):
    blueprint_class = Blueprint

    @classmethod
    def inherit(cls, parent_api, *args, **kwargs):
        api = cls(*args, **kwargs)
        api.mixin(parent_api)
        api.services.update(parent_api.services)
        return api

    def __init__(self, version, import_name, url_prefix=None, spec_builder=None):
        if url_prefix is None:
            url_prefix = '/api/%s' % version
        super(ApiVersion, self).__init__('', url_prefix)
        self.version = version
        self.import_name = import_name
        self.services = {}
        self.spec_builder = spec_builder or build_swagger_spec

    def build_apispec(self, refresh=False, **kwargs):
        if 'apispec' not in self.__dict__ or refresh:
            self.__dict__['apispec'] = self.spec_builder(self, **kwargs).to_dict()
        return self.__dict__['apispec']

    def add_service(self, service):
        self.services[service.name] = service

    def iter_services(self):
        return self.services.values()

    def service(self, name, **kwargs):
        service = ApiService(name, **kwargs)
        self.add_service(service)
        return service

    def __getattr__(self, name):
        if name in self.services:
            return self.services[name]
        raise ValueError("Service '%s' does not exist in API %s" % (name, self.version))

    def _merge_services_endpoints(self):
        rules = list(self.rules)
        endpoint_funcs = dict(self.endpoint_funcs)
        for service in self.services.values():
            for rule, endpoint, func, options in service.iter_endpoints():
                endpoint = service.name + '_' + endpoint
                rules.append((join_url_rule(service.url_prefix, rule), endpoint, options))
                endpoint_funcs[endpoint] = func
        return rules, endpoint_funcs

    def iter_all_endpoints(self):
        rules, endpoint_funcs = self._merge_services_endpoints()
        for rule, endpoint, options in rules:
            yield rule, endpoint, endpoint_funcs[endpoint], options

    def as_blueprint(self, name=None, client_var_name="API", client_class_name="SwaggerClient", spec_builder_options=None):
        if not spec_builder_options:
            spec_builder_options = {}
        if not name:
            name = "api_%s" % self.version.replace('.', '_')
        bp = self.blueprint_class(name, self.import_name, url_prefix=self.url_prefix)

        @bp.before_request
        def before_request():
            g.is_api_call = True

        @bp.after_request
        def after_request(resp):
            resp.headers.add('Cache-Control', 'no-cache')
            resp.headers.add('X-Api-Version', self.version)
            return resp

        endpoint_funcs = {}
        for rule, endpoint, func, options in self.iter_all_endpoints():
            endpoint_funcs.setdefault(endpoint, wrap_service_func_for_endpoint(func))
            bp.add_url_rule(rule, endpoint, endpoint_funcs[endpoint], **options)

        @bp.route('/spec.json')
        def get_spec():
            return jsonify(**self.build_apispec(**spec_builder_options))

        @bp.route("/client.js")
        def get_client():
            return build_swagger_client(self.build_apispec(**spec_builder_options),
                url_for('.get_spec', _external=True), client_var_name, client_class_name)

        return bp



from frasco.models import db


class InvoiceModelMixin(object):
    ref = db.Column(db.String, index=True)
    currency = db.Column(db.String)
    subtotal = db.Column(db.Float)
    total = db.Column(db.Float)
    tax_rate = db.Column(db.Float)
    tax_amount = db.Column(db.Float)
    description = db.Column(db.String)
    name = db.Column(db.String)
    email = db.Column(db.String)
    address_line1 = db.Column(db.String)
    address_line2 = db.Column(db.String)
    address_city = db.Column(db.String)
    address_state = db.Column(db.String)
    address_zip = db.Column(db.String)
    address_country = db.Column(db.String)
    country = db.Column(db.String)
    customer_special_mention = db.Column(db.String)
    issued_at = db.Column(db.DateTime)
    charge_id = db.Column(db.String)
    external_id = db.Column(db.String)


class InvoiceItemModelMixin(object):
    amount = db.Column(db.Float)
    description = db.Column(db.String)
    quantity = db.Column(db.Integer)
    subtotal = db.Column(db.Float)
    currency = db.Column(db.String)
    external_id = db.Column(db.String)



from frasco import lazy_translate, request, flash, redirect
from frasco.ext import *
from frasco.utils import unknown_value
from frasco.models import as_transaction
from frasco.mail import send_mail
import stripe
import datetime
import time
import os
import json

from frasco.billing.eu_vat import get_exchange_rate, is_eu_country, model_rate_updated
from frasco.billing.invoicing import send_failed_invoice_mail

from .signals import *
from .webhook import webhook_blueprint
from .invoice import *
from .model import *


stripe.enable_telemetry = False


class FrascoStripe(Extension):
    name = "frasco_stripe"
    defaults = {"default_currency": None,
                "default_plan": None,
                "no_payment_redirect_to": None,
                "no_payment_message": None,
                "subscription_past_due_message": lazy_translate(
                    u"We attempted to charge your credit card for your subscription but it failed."
                    "Please check your credit card details"),
                "debug_trial_period": None,
                "send_trial_will_end_email": True,
                "send_failed_invoice_mail": True,
                "invoice_ref_kwargs": {},
                "webhook_validate_event": False,
                "default_subscription_tax_percent": None}

    def _init_app(self, app, state):
        stripe.api_key = state.require_option('api_key')
        state.Model = state.import_option('model')
        state.subscription_enabled = hasattr(state.Model, 'stripe_subscription_id')

        app.register_blueprint(webhook_blueprint)

        if has_extension("frasco_mail", app):
            app.extensions.frasco_mail.add_templates_from_package(__name__)

        if has_extension('frasco_eu_vat', app):
            def on_model_rate_updated(sender):
                if sender.stripe_subscription:
                    sender.stripe_subscription.tax_percent = sender.eu_vat_rate
                    sender.stripe_subscription.save()
            model_rate_updated.connect(on_model_rate_updated, weak=True)

        stripe_event_signal('customer_source_updated').connect(on_source_event)
        stripe_event_signal('customer_source_deleted').connect(on_source_event)
        stripe_event_signal('invoice_payment_succeeded').connect(on_invoice_payment)
        stripe_event_signal('invoice_payment_failed').connect(on_invoice_payment)
        if state.subscription_enabled:
            stripe_event_signal('customer_subscription_updated').connect(on_subscription_event)
            stripe_event_signal('customer_subscription_deleted').connect(on_subscription_event)
            stripe_event_signal('customer_subscription_trial_will_end').connect(on_trial_will_end)


def check_stripe_subscription_is_valid(obj):
    state = get_extension_state('frasco_stripe')
    if obj.is_stripe_subscription_valid():
        if state.options['no_payment_message']:
            flash(state.options['no_payment_message'], 'error')
        if state.options['no_payment_redirect_to']:
            return redirect(state.options['no_payment_redirect_to'])
    if obj.plan_status == 'past_due' and state.options['subscription_past_due_message']:
        flash(state.options['subscription_past_due_message'], 'error')


@as_transaction
def on_source_event(sender, stripe_event):
    state = get_extension_state('frasco_stripe')
    source = stripe_event.data.object
    obj = state.Model.query_by_stripe_customer(source.customer).first()
    if not obj:
        return
    obj._update_stripe_source()


@as_transaction
def on_subscription_event(sender, stripe_event):
    state = get_extension_state('frasco_stripe')
    subscription = stripe_event.data.object
    obj = state.Model.query_by_stripe_customer(subscription.customer).first()
    if not obj:
        return
    obj._update_stripe_subscription(subscription)


@as_transaction
def on_trial_will_end(sender, stripe_event):
    state = get_extension_state('frasco_stripe')
    subscription = stripe_event.data.object
    obj = state.Model.query_by_stripe_subscription(subscription.id).first()
    if not obj:
        return
    if state.options['send_trial_will_end_email']:
        send_mail(getattr(obj, obj.__stripe_email_property__), 'stripe/trial_will_end.txt', obj=obj)


@as_transaction
def on_invoice_payment(sender, stripe_event):
    state = get_extension_state('frasco_stripe')
    invoice = stripe_event.data.object
    if not invoice.customer:
        return

    obj = state.Model.query_by_stripe_customer(invoice.customer).first()
    if not obj or invoice.total == 0:
        return

    if invoice.subscription:
        sub_obj = None
        if obj.stripe_subscription_id == invoice.subscription:
            sub_obj = obj
        else:
            sub_obj = state.Model.query_by_stripe_subscription(invoice.subscription).first()
        if sub_obj:
            sub_obj.plan_status = sub_obj.stripe_subscription.status
            sub_obj.update_last_stripe_subscription_charge(invoice)

    if getattr(obj, '__stripe_has_eu_vat__', False) and getattr(obj, '__stripe_has_billing_fields__', False) and is_eu_country(obj.billing_country):
        invoice.metadata['eu_vat_exchange_rate'] = get_exchange_rate(obj.billing_country, invoice.currency.upper())
        if invoice.tax:
            invoice.metadata['eu_vat_amount'] = round(invoice.tax * invoice.metadata['eu_vat_exchange_rate'])
        if obj.eu_vat_number:
            invoice.metadata['eu_vat_number'] = obj.eu_vat_number
        invoice.save()

    invoice_payment.send(invoice)

    if has_extension('frasco_invoicing'):
        if invoice.paid:
            create_invoice_from_stripe(obj, invoice)
        elif not invoice.paid and state.options['send_failed_invoice_mail']:
            send_failed_invoice_mail(getattr(obj, obj.__stripe_email_property__), invoice)



from flask import g, current_app
from frasco.ext import get_extension_state
from frasco.utils import remove_yaml_frontmatter
from flask_mail import Message, Attachment
from jinja2 import TemplateNotFound
import yaml
import os
import markdown
import premailer
import re
import datetime
import logging

try:
    # html2text being licensed under the GPL, it is not
    # directly included and may be not available
    import html2text
except ImportError:
    html2text = None


__all__ = ('render_message', 'create_message', 'log_message')


def get_template_source(template_filename, locale=None):
    state = get_extension_state('frasco_mail')
    filename, ext = os.path.splitext(template_filename)

    localized_filename = None
    if state.options['localized_emails']:
        if locale is None and 'current_locale' in g:
            locale = g.current_locale
        if locale and locale != state.options['default_locale']:
            localized_filename = state.options['localized_emails'].format(**{
                "locale": locale, "filename": template_filename,
                "name": filename, "ext": ext})

    source = None
    for tpl_filename in [localized_filename, template_filename]:
        if not tpl_filename:
            continue
        filename, ext = os.path.splitext(tpl_filename)
        auto_ext = not ext
        if auto_ext:
            ext = ".md,html,txt"
        for source_filename in ["%s.%s" % (filename, e) for e in ext[1:].split(",")]:
            try:
                source, _, __ = state.jinja_env.loader.get_source(state.jinja_env, source_filename)
            except TemplateNotFound:
                continue
            if source:
                return tpl_filename, source

    raise TemplateNotFound(template_filename)


def render_message(template_filename, vars=None, auto_render_missing_content_type=None,
                   auto_html_layout="layouts/text.html", auto_markdown_template="layouts/markdown.html",
                   **kwargs_vars):

    state = get_extension_state('frasco_mail')
    text_body = None
    html_body = None
    vars = vars or {}
    vars.update(kwargs_vars)

    template_filename, source = get_template_source(template_filename, vars.get('locale'))
    source, frontmatter_source = remove_yaml_frontmatter(source, True)
    frontmatter = None
    if frontmatter_source:
        frontmatter = yaml.safe_load(re.sub("^---\n", "", frontmatter_source))
        if frontmatter:
            for k, v in frontmatter.items():
                vars.setdefault(k, state.jinja_env.from_string(v).render(**vars))

    filename, ext = os.path.splitext(template_filename)
    auto_ext = not ext
    if auto_ext:
        ext = ".md,html,txt"
    templates = [("%s.%s" % (filename, e), e) for e in ext[1:].split(",")]
    for tpl_filename, ext in templates:
        try:
            rendered = state.jinja_env.get_template(tpl_filename).render(**vars)
        except TemplateNotFound:
            if not auto_ext:
                raise
            continue
        if ext == "html":
            html_body = rendered
        elif ext == "txt":
            text_body = rendered
        elif ext == "md":
            text_body = rendered
            content = markdown.markdown(rendered, **state.options["markdown_options"])
            html_body = state.jinja_env.get_template(auto_markdown_template).render(
                content=content, **vars)
        if auto_ext and html_body and text_body:
            break

    if (auto_render_missing_content_type is not None and auto_render_missing_content_type) or \
        (auto_render_missing_content_type is None and state.options["auto_render_missing_content_type"]):
        if html_body is None:
            html_body = state.jinja_env.get_template(auto_html_layout).render(
                text_body=text_body, **vars)
        if text_body is None and html2text is not None:
            text_body = html2text.html2text(html_body)

    if html_body and state.options["inline_css"]:
        html_body = premailer.transform(html_body)

    return (frontmatter, vars, text_body, html_body)


def create_message(to, tpl, **kwargs):
    recipients = to if isinstance(to, (list, tuple)) else [to]
    frontmatter, vars, text_body, html_body = render_message(tpl, **kwargs)

    kwargs = {}
    for k in ('subject', 'sender', 'cc', 'bcc', 'attachments', 'reply_to', 'send_date', 'charset', 'extra_headers'):
        if k in vars:
            kwargs[k] = vars[k]
        elif frontmatter and k in frontmatter:
            kwargs[k] = frontmatter[k]
    kwargs["date"] = kwargs.pop("send_date", None)

    if not kwargs.get("subject"):
        raise ValueError("Missing subject for email with template '%s'" % tpl)
    subject = kwargs.pop("subject")
    attachments = kwargs.pop("attachments", None)

    msg = Message(recipients=recipients, subject=subject, body=text_body, html=html_body, **kwargs)
    msg.template = tpl

    if attachments:
        for attachment in attachments:
            if isinstance(attachment, Attachment):
                msg.attachments.append(attachment)
            elif isinstance(attachment, dict):
                msg.attach(**attachment)
            else:
                msg.attach(attachment)

    return msg


def log_message(message, dump_dir=None):
    logging.getLogger('frasco.mail').info("Email %s sent to %s from %s titled \"%s\"" % (
        getattr(message, 'template', '[no template]'), message.recipients, message.sender, message.subject))
    if dump_dir:
        if not os.path.exists(dump_dir):
            os.mkdir(dump_dir, 0777)
        filename = os.path.join(dump_dir, "-".join([
            datetime.datetime.now().isoformat("-"),
            os.path.splitext(getattr(message, 'template', ''))[0].replace("/", "_"),
                "-".join(message.recipients)]))
        if message.body:
            with open(filename + ".txt", "w") as f:
                f.write(message.body.encode('utf-8'))
        if message.html:
            with open(filename + ".html", "w") as f:
                f.write(message.html.encode('utf-8'))


_url_regexp = re.compile(r'(http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\(\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+)')
def clickable_links(text):
    return _url_regexp.sub(r'<a href="\1">\1</a>', text)
